// =========================================
// The special Minisat constraint that makes
// Sure that the candidate automaton does not
// accept a language too small.
// =========================================

#ifndef ACCEPTENOUGHCONSTRAINT_H
#define ACCEPTENOUGHCONSTRAINT_H

#include "Minisat/Solver.h"
#include "NBW.hpp"
#include "MemoryRecycler.hpp"
#include <tr1/unordered_map>

#ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
#ifndef __x86_64
typedef uint64_t NotTooMuchCauseTrackingType;
#else
typedef __uint128_t NotTooMuchCauseTrackingType;
#endif
#endif

// #undef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT

//=====================================
// The main constraint class
//=====================================
class AcceptEnoughConstraint : public Constr
{
private:
    Solver &solver;
    const NBW &pos;
    uint targetNofStates;
    Var ***transitionVars;
    Var *acceptanceVars;
    uint chosenBound;

    #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
    NotTooMuchCauseTrackingType lastCause;
    #endif

    MemoryRecycler<int> memoryRecycler; // Used for state vectors

    #ifdef TRACK_CAUSE_IN_NOT_TOO_MUCH_CONSTRAINT
    std::vector<boost::tuple<int*,uint,NotTooMuchCauseTrackingType > > todoList; // CandidateStates. First one must be int* as the values might become negative temporarily
    #else
    std::vector<boost::tuple<int*,uint > > todoList; // CandidateStates. First one must be int* as the values might become negative temporarily
    #endif

    // The actual working function
    bool performCheck(uint bound);

public:
    AcceptEnoughConstraint(Solver& _S, const NBW &_pos, int _targetNofStates, Var ***_transitionVars, Var *_acceptanceVars, uint _bound);

    // Minisat functions
    virtual void remove    (Solver& s, bool just_dealloc = false);
    virtual bool propagate (Solver& S, Lit p, bool& keep_watch);    // ('keep_watch' is set to FALSE beftore call to this method)
    virtual void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
    virtual void undo      (Solver& UNUSED(S), Lit UNUSED(p)) { };
};

#endif // ACCEPTENOUGHCONSTRAINT_H
