// =========================================
// The special Minisat clause that makes
// Sure that the candidate automaton does not
// accept a language too large.
// =========================================
#ifndef ACCEPTNOTTOOMUCHCONSTRAINT_H
#define ACCEPTNOTTOOMUCHCONSTRAINT_H

#include "Minisat/Solver.h"
#include "NBW.hpp"
#include <tr1/unordered_map>
#include <tr1/unordered_set>
#include <stdint.h>

class AcceptNotTooMuchConstraint : public Constr
{
private:
    Solver &solver;
    const NBW &neg;
    uint targetNofStates;
    Var ***transitionVars;
    Var *acceptanceVars;

    // The last reason, to be used by "calcReason"
    std::set<Lit> lastReason;

    // The double-DFS reason tables and graph tables.
    int *dfsVisitedMode; // States which nodes in the product graph have already been visited. (Mode 0 = not visited, Mode 1 = visited in first round, Mode 2 = visited in second round)
    Lit *reasons1; // For tracking the reason why a node is reachable in the graph
    Lit *reasons2; // For tracking the reason why a node is reachable in the graph
    int *prevReasonPointers; // For tracking which predecessor was used for building the initial graph
    std::vector<uint> acceptingStatesInPostOrder;

    // This function computes the position in one of the above four tables for a given state in the product graph.
    inline int posInTable(uint referenceAutNr, uint candidateAutNr, uint mode) const { return mode+3*(referenceAutNr*targetNofStates+candidateAutNr); }

    // Internal functions. Perform the main part of the computation
    void dfsFirstRound(uint currentNode);
    bool dfsSecondRound(uint basisNode, uint currentNode);


public:
    AcceptNotTooMuchConstraint(Solver &_solver, const NBW &_neg, uint _targetNofStates, Var ***_transitionVars, Var *_acceptanceVars);
    ~AcceptNotTooMuchConstraint();

    // Minisat functions
    virtual void remove    (Solver& s, bool just_dealloc = false);
    virtual bool propagate (Solver& S, Lit p, bool& keep_watch);    // ('keep_watch' is set to FALSE beftore call to this method)
    virtual bool simplify  (Solver& UNUSED(S)) { return false; }
    virtual void undo      (Solver& UNUSED(S), Lit UNUSED(p)) { }
    virtual void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);

};

#endif // ACCEPTNOTTOOMUCHCLAUSE_H
