#ifndef MEMORYRECYCLER_H
#define MEMORYRECYCLER_H

#include <vector>

// A class for recycling memory in T-Arrays
// in order to reduce new & free-calls

template<typename T> class MemoryRecycler {
private:
    int arraySize;
    std::vector<T*> allocList;
    uint currentPosInAllocList;
public:
    MemoryRecycler(int _arraySize) : arraySize(_arraySize) {}
    ~MemoryRecycler() {
        typedef typename std::vector<T*>::iterator IteratorType;
        for (IteratorType it = allocList.begin();it!=allocList.end();it++) delete[] *it;
    }
    inline void clear() { currentPosInAllocList = 0; }
    inline T* getNew() {
        if (allocList.size()==currentPosInAllocList) {
            allocList.push_back(new T[arraySize]);
        }
        return allocList[currentPosInAllocList++];
    }
};

#endif // MEMORYRECYCLER_H
