//========================================================
// A class storing a non-determinstic Büchi word automaton
//========================================================

#include "NBW.hpp"
#include <fstream>
#include <map>
#include <algorithm>

//========================================================
// PreNBW functions
//========================================================
PreNBW::PreNBW(const char *inputNeverClaimFilename) {
#define WASTE_UNTIL_END_OF_LINE(a) { std::string waste; std::getline((a),waste); }

        // First run: Get all the states
        {
                std::ifstream in(inputNeverClaimFilename);
                if (in.fail()) {
                    std::cerr << "Cannot open: " << inputNeverClaimFilename << "\n";
                    throw GeneralException("Failed to read input NBW.");
                }
                try {
                        expect(in,"never { /* ");
                        bool finished = false;

                        WASTE_UNTIL_END_OF_LINE(in)

                        while (!finished) {
                                // Labels am Start
                                char stateLabel[256];
                                in.get(stateLabel,256,':');
                                std::string stateName = stateLabel;
                                boost::trim(stateName);
                                if (stateName!="}") {
                                        expect(in,":");

                                        bool accept = (stateName.substr(0,6)=="accept");
                                        states.insert(stateName);
                                        if (accept) acceptingstates.insert(stateName);
                                        if (stateName.substr(stateName.length()-4,4)=="init") initialState = stateName;

                                        // Ignore the rest of the state
                                        std::string waste;
                                        std::getline(in,waste);
                                        boost::trim(waste);
                                        while ((!(waste=="fi;")) && (!(waste=="skip"))) {
                                                std::getline(in,waste);
                                                boost::trim(waste);
                                        }
                                } else {
                                        // we're done!
                                        finished = true;
                                }
                        }
                } catch (GeneralException e) {
                        throw GeneralException("LTL compilation error. Reason: "+e.getMessage());
                }
        }

        // Second: Now get the transitions
        // First run: Get all the states
        {
                std::ifstream in(inputNeverClaimFilename);
                try {
                        expect(in,"never { /* ");
                        bool finished = false;

                        WASTE_UNTIL_END_OF_LINE(in);

                        while (!finished) {
                                // Labels am Start
                                char stateLabel[256];
                                in.get(stateLabel,256,':');
                                std::string stateName = stateLabel;
                                boost::trim(stateName);
                                // std::cout << "StateLabel: " << stateName << "\n";
                                if (stateName!="}") {
                                        expect(in,":");

                                        WASTE_UNTIL_END_OF_LINE(in);

                                        // Assert: if
                                        std::string ifString;
                                        std::getline(in,ifString);
                                        boost::trim(ifString);
                                        if (ifString=="if") {

                                                // Read the connections
                                                std::string nextTransition;
                                                std::getline(in,nextTransition);
                                                boost::trim(nextTransition);
                                                while (nextTransition!="fi;") {

                                                        if (nextTransition.substr(0,3)!=":: ")
                                                                throw GeneralException("LTL compilation error: Illegal transition line:"+nextTransition);
                                                        size_t endIndex = nextTransition.find(" -> goto ",0);
                                                        if (endIndex == std::string::npos) {
                                                                throw GeneralException("LTL compilation error: Cannot find '-> goto' in transition: "+nextTransition);
                                                        }

                                                        std::string condition = nextTransition.substr(0,endIndex).substr(3);
                                                        std::string to = nextTransition.substr(endIndex+9);
                                                        boost::trim(to);
                                                        addSPINNotationTransition(stateName,to,condition);

                                                        // Next line
                                                        std::getline(in,nextTransition);
                                                        boost::trim(nextTransition);
                                                }
                                        } else if (ifString=="skip") {
                                                addSPINNotationTransition(stateName,stateName,"1");
                                        } else if (ifString=="false;") {
                                                throw "Not implemented!";
                                        } else {
                                                throw GeneralException("LTL compilation error: 'if' expected but "+ifString+" found.");
                                        }
                                } else {
                                        finished = true;
                                }
                        }
                } catch (GeneralException e) {
                        throw GeneralException("LTL compilation error. Reason: "+e.getMessage());
                }
        }

    if (initialState=="") throw GeneralException("Error reading NBW: No initial state given!");
}


void PreNBW::addSPINNotationTransition(std::string from, std::string to, std::string what) {

        // 1. Check if there are braces around the disjunction (as made by SPOT) and remove them
        //std::cerr << "Check: " << what << "\n";
        int braceLevel = 0;
        bool foundDisjunctionInBraces = false;
        for (uint i=0;i<what.length();i++) {
                switch (what[i]) {
                        case '(':
                                braceLevel += 1;
                                break;
                        case ')':
                                braceLevel -= 1;
                                break;
                        case '|':
                                if (braceLevel>0) foundDisjunctionInBraces = true;
                                break;
                }
                if (braceLevel<0) throw "Error: Unmatched braces in LTLCompiler::convertSPINBoolFormulaToPrefixNotation";
        }
        if (braceLevel!=0) throw "Error: Unmatched braces in LTLCompiler::convertSPINBoolFormulaToPrefixNotation";
        if (foundDisjunctionInBraces) {
                if ((what[0]!='(') || (what[what.length()-1]!=')')) throw "Error: Too complex braces in LTLCompiler::convertSPINBoolFormulaToPrefixNotation";
                what = what.substr(1,what.length()-2);
                // std::cerr << "Newcheck: " << what << "\n";
        }

        // Explode agains "||" operators
        std::vector<std::string> orParts;
        boost::iter_split(orParts, what, boost::first_finder(" || "));

        for (uint i=0;i<orParts.size();i++) {

                std::string current = orParts[i];
                if (current[0]=='(') {
                        assert(current[current.length()-1]==')');
                        current = current.substr(1,current.length()-2);
                } else {
                        assert(current[current.length()-1]!=')');
                }

                std::vector<std::string> andParts;
                boost::iter_split(andParts, current, boost::first_finder(" && "));

                std::vector<std::string> conjuncts;

                for (uint j=0;j<andParts.size();j++) {

                        std::string current = andParts[j];
                        //std::cout << "Current: " << current << std::endl;
                        bool negated = current[0]=='!';
                        std::string negatedString = "";
                        if (negated) {
                                current = current.substr(1);
                                negatedString = "~";
                        }

                        if (current=="1") {
                                conjuncts.push_back(negatedString+"True");
                        } else if (current=="safe") {
                                conjuncts.push_back(negatedString+"safe");
                                alphabet.insert("safe");
                        } else {
                                //assert(current[0]=='v');
                                /*istringstream tonumber(current.substr(1));
                                uint number;
                                tonumber >> number;
                                if (tonumber.fail()) throw GeneralException("LTL compilation error: Illegal variable: "+current);*/
                                conjuncts.push_back(negatedString+current);
                                alphabet.insert(current);
                        }
                }

                transitions.insert(boost::make_tuple(from,conjuncts,to));
        }

}

void PreNBW::addToAlphabet(std::set<std::string> stuff) {
    for (std::set<std::string>::iterator it = stuff.begin();it!=stuff.end();it++) {
        alphabet.insert(*it);
    }
}

//========================================================
// NBW functions
//========================================================
NBW::NBW(uint _nofStates, uint _sizeSigma) : nofStates(_nofStates), sizeSigma(_sizeSigma) {
    accepting = new bool[nofStates];
    transitions = new bool**[nofStates];
    for (uint i=0;i<nofStates;i++) {
        accepting[i] = false;
        transitions[i] = new bool*[sizeSigma];
        for (uint j=0;j<sizeSigma;j++) {
            transitions[i][j] = new bool[nofStates];
            for (uint k=0;k<nofStates;k++) transitions[i][j][k] = false;
        }
    }
}


NBW::NBW(const PreNBW &source) {

    // Copy basic structure of other NBW
    nofStates = source.states.size();
    sizeSigma = 1 << source.alphabet.size();
    if (source.alphabet.size()>15) throw GeneralException("Input alphabet is too large - the automaton instance cannot be handled by this tool.");
    accepting = new bool[nofStates];
    transitions = new bool**[nofStates];
    for (uint i=0;i<nofStates;i++) {
        accepting[i] = false;
        transitions[i] = new bool*[sizeSigma];
        for (uint j=0;j<sizeSigma;j++) {
            transitions[i][j] = new bool[nofStates];
            for (uint k=0;k<nofStates;k++) transitions[i][j][k] = false;
        }
    }

    // Make a state list
    std::map<std::string,uint> stateNrs;
    stateNrs[source.initialState] = 0;
    for (std::set<std::string>::iterator it = source.states.begin();it!=source.states.end();it++) {
        if ((*it) != source.initialState) {
            int next = stateNrs.size();
            stateNrs[*it] = next;
        }
    }

    // Copy accepting states
    for (std::set<std::string>::iterator it = source.acceptingstates.begin();it!=source.acceptingstates.end();it++) {
        int pos = stateNrs.at(*it);
        accepting[pos] = true;
    }

    // Make alphabet list
    std::vector<std::string> alphabetList;
    std::map<std::string,uint> alphabetListPos;
    for (std::set<std::string>::iterator it = source.alphabet.begin();it!=source.alphabet.end();it++) {
        alphabetList.push_back(*it);
        int next = alphabetListPos.size();
        alphabetListPos[*it] = next;
    }

    for (uint i=0;i<sizeSigma;i++) {
        std::ostringstream os;
        for (uint j=0;j<source.alphabet.size();j++) {
            if (j!=0) os << " && ";
            if ((i & (1 << j))==0) os << " !";
            os << alphabetList[j];
        }
        alphabet.push_back(os.str());
    }

    // Get going on the transitions
    for (std::set<boost::tuple<std::string,std::vector<std::string>,std::string> >::iterator it = source.transitions.begin();it!=source.transitions.end();it++) {
        boost::tuple<std::string,std::vector<std::string>,std::string> current = *it;

        // Compute transition mask
        uint mask = 0;
        uint match = 0;
        std::vector<std::string> &parts = current.get<1>();
        for (uint i=0;i<parts.size();i++) {
            if (parts[i]=="True") {
                // Do nothing
            } else if (parts[i][0]=='~') {
                std::string rest = parts[i].substr(1,std::string::npos);
                uint pos = alphabetListPos.at(rest);
                mask |= (1 << pos);
            } else {
                std::string rest = parts[i];
                uint pos = alphabetListPos.at(rest);
                mask |= (1 << pos);
                match |= (1 << pos);
            }
        }

        int nrFrom = stateNrs.at(current.get<0>());
        int nrTo = stateNrs.at(current.get<2>());
        // std::cout << "Candidate: " << nrFrom << " to: " << nrTo << " via: " << mask << "," << match << "\n";

        for (uint i=0;i<sizeSigma;i++) {
            if ((i & mask)==match) {
                transitions[nrFrom][i][nrTo] = true;
                // std::cout << "From: " << nrFrom << " to: " << nrTo << " via: " << i << "\n";
            }
        }

    }
}

NBW::~NBW() {
    delete [] accepting;
    for (uint i=0;i<nofStates;i++) {
        for (uint j=0;j<sizeSigma;j++) {
            delete[] transitions[i][j];
        }
        delete[] transitions[i];
    }
    delete[] transitions;
}

void NBW::print() {

    // Build state names vector
    std::vector<std::string> stateNames;
    for (uint i=0;i<nofStates;i++) {
        std::ostringstream os;
        if (accepting[i]) os << "accept_";
        os << "T";
        os << i;
        if (i==0) os << "_init";
        stateNames.push_back(os.str());

    }

    std::cout << "never { /* Formula not given */\n";
    for (uint i=0;i<nofStates;i++) {
        std::cout << stateNames[i] << ":\n\tif\n";
        for (uint j=0;j<sizeSigma;j++) {
            for (uint k=0;k<nofStates;k++) {
                if (transitions[i][j][k]) {
                    std::cout << "\t:: (" << alphabet[j] << ") -> goto " << stateNames[k] << ";\n";
                }
            }
        }
        std::cout << "\tfi;\n";
    }
    std::cout << "};\n";

}
