#include "ReferenceAutomatonBitMaskTypeTrie.hpp"

//=====================================
// A class for storing the nodes explored
// in the graph so far - Requires sizeof(ReferenceAutomatonBitMaskType)==sizeof(uint)
// and graceful initialisation of the type T with (uint)0
//
// (C) 2010 by Ruediger Ehlers
//=====================================

ReferenceAutomatonBitMaskTypeTrie::ReferenceAutomatonBitMaskTypeTrie(int _depth, int _branching) : depth(_depth), branching(_branching) {
    root = 0;
}

size_t** ReferenceAutomatonBitMaskTypeTrie::find_recurse(size_t **ptr, int *value, int pos) {
    if (pos==depth) {
        return ptr;
    }
    // We have not reached the bottom yet...
    if (*ptr==0) {
        *ptr = reinterpret_cast<size_t*>(new size_t*[branching]);
        for (int i=0;i<branching;i++) (*ptr)[i] = 0;
    }
    assert(pos<depth);
    assert(pos>=0);
    assert(value[pos]>=0);
    assert(value[pos]<branching);
    return find_recurse(reinterpret_cast<size_t**>(&((*ptr)[value[pos]])),value,pos+1);
}

MaskReference ReferenceAutomatonBitMaskTypeTrie::find(int *value) {
    return MaskReference(reinterpret_cast<size_t*>(find_recurse(&root, value, 0)));
}

/*void ReferenceAutomatonBitMaskTypeTrie::add(int *value, const T &data) {
    ReferenceAutomatonBitMaskTypeTrie* ptr = find(value);
    if (*ptr!=0) delete *ptr;
    *ptr = T(data);
}*/

ReferenceAutomatonBitMaskTypeTrie::~ReferenceAutomatonBitMaskTypeTrie() {
    delete_recurse(&root,0);
}

void ReferenceAutomatonBitMaskTypeTrie::delete_recurse(size_t **ptr, int pos) {
    if (*ptr!=NULL) {
        if (pos==depth) {
            #ifndef NO_SUPPORT_FOR_LARGE_NBAS
            delete reinterpret_cast<std::tr1::unordered_set<size_t> *>(*ptr);
            #endif
        } else {
            for (int i=0;i<branching;i++) {
                delete_recurse(reinterpret_cast<size_t**>(&((*ptr)[i])),pos+1);
            }
            delete[] *ptr;
        }
    }
}
