#ifndef INTTRIE_H
#define INTTRIE_H

#include <stdlib.h>
#include <assert.h>
#include <tr1/unordered_set>

//=====================================
// This type is used as a std::set<..> for states in the reference automaton.
//=====================================
#ifdef NO_SUPPORT_FOR_LARGE_NBAS
class MaskReference {
private:
    size_t *mask;
public:
    MaskReference(size_t *where) : mask(where) {}
    void insert(size_t value) { *mask |= (1 << value); }
    bool isSet(size_t value) { return ((*mask) & (1 << value))!=0; }
};
#else
class MaskReference {
private:
    size_t *mask;
public:
    inline MaskReference(size_t *where) : mask(where) {
        if (*mask==0) { *mask = reinterpret_cast<size_t>(new std::tr1::unordered_set<size_t>()); };
    }
    inline void insert(size_t value) { reinterpret_cast<std::tr1::unordered_set<size_t> *>(*mask)->insert(value); }
    inline bool isSet(size_t value) { return reinterpret_cast<std::tr1::unordered_set<size_t> *>(*mask)->count(value)>0; }
};
#endif

//=====================================
// A class for storing the nodes explored
// in the graph so far - Requires sizeof(ReferenceAutomatonBitMaskType)==sizeof(uint)
// and graceful initialisation of the type T with (uint)0
//
// (C) 2010 by Ruediger Ehlers
//=====================================
class ReferenceAutomatonBitMaskTypeTrie {
private:
    int depth;
    int branching;
    size_t *root;

    size_t **find_recurse(size_t **ptr, int *value, int pos);
    void delete_recurse(size_t **ptr, int pos);

public:
    ReferenceAutomatonBitMaskTypeTrie(int _depth, int _branching);
    ~ReferenceAutomatonBitMaskTypeTrie();
    //void add(int *value, const ReferenceAutomatonBitMaskTypeTrie &data);
    MaskReference find(int *value);
};

#endif // INTTRIE_H
