// =========================================
// The special Minisat constraint that breaks
// symmetry partially
// =========================================

#ifndef SYMMETRYBREAKER_H
#define SYMMETRYBREAKER_H

#include "Minisat/Solver.h"
#include "NBW.hpp"
#include "MemoryRecycler.hpp"
#include <tr1/unordered_set>


class LexicographicConstraint : public Constr
{
private:
    Var *equality1;
    Var *equality2;
    Lit smaller1;
    Lit smaller2;
    bool *equalities;
    int nofEqualities;
    int *posMapper;
    int size; // Of the equality part

public:
    LexicographicConstraint(Solver& S, Var *_equality1, Var *_equality2, Lit _smaller1, Lit smaller2, int size); // Takes ownership of the arrays
    ~LexicographicConstraint();

    // Minisat functions
    virtual void remove    (Solver& s, bool just_dealloc = false);
    virtual bool propagate (Solver& S, Lit p, bool& keep_watch);    // ('keep_watch' is set to FALSE beftore call to this method)
    virtual void calcReason(Solver& S, Lit p, vec<Lit>& out_reason);
    virtual void undo      (Solver& UNUSED(S), Lit UNUSED(p));
};


//=====================================
// The main constraint class
//
// Symmetry breaking here is based on partial
// lexicographic comparisons
//
// For no 1<=i<N, exchanging state i with i+1 must result
// in a lexicographically smaller automaton (for some
// ordering)
//=====================================
void make_symmetry_breaking_constraints(Solver& _S, uint _targetNofStates, uint sizeSigma, Var ***_transitionVars, Var *_acceptanceVars, uint whichState);

#endif // ACCEPTENOUGHCONSTRAINT_H
