#! /usr/bin/env python
# -*- coding: ISO-8859-15 -*-
# 
# A wrapper for performing the task of minimizing non-deterministic Buechi automata
#
# (C) 2010 by Ruediger Ehlers
# Free for academic use

import getopt
import sys
import tempfile
import shutil
import os
import subprocess
import time
import signal

# Usage
usage = "NBW-minimize v.0.1 - (C) 2010 by Ruediger Ehlers\n\nUsage: minimize.py [--timeout <nofSeconds>] directory\n\nIn the directory specified, there must be two automata, named 'pos.txt' and 'neg.txt', where the former represents the automaton to minimize, whereas the latter needs to represent the negated language (otherwise the minimizer is unsound). Both automata need to be given as SPIN never claims. If the computation time of some call to the SAT solver exceeds the given number of seconds, the process is aborted and the best found automaton is returned to the standard output.\n"

# Parse options
args = sys.argv[1:]
basePath = sys.argv[0].rsplit(os.sep,1)[0]
timeout = -1
try:
    (optlist, args) = getopt.getopt(args, 'x', ['timeout='])
except getopt.GetoptError:
    print usage
    exit(1)
if (len(args)<>1) or (len(optlist)>1):
    print usage
    exit(1)
if len(optlist)==1:
    try:
        timeout = float(optlist[0][1])
    except ValueError:
        print usage
        exit(1)

# Functions
def getNofStatesInSpinNeverClaimFile(infile):
    nofStatesSoFar = 0
    with open(infile,'r') as myfile:
        for line in myfile.readlines():
            if (line.endswith(":\n")):
                nofStatesSoFar += 1
    return nofStatesSoFar

def fileContainsValidNeverClaim(infile):
    with open(infile,'r') as myfile:
        for line in myfile.readlines():
            if (line.strip()=="}"):
                return True
    return False


# Everything OK? Then let's go!
try:
    if (not fileContainsValidNeverClaim(args[0]+os.sep+"pos.txt")):
        print "Error: The file 'pos.txt' in the given directory does not contain a valid never claim."
    if (not fileContainsValidNeverClaim(args[0]+os.sep+"neg.txt")):
        print "Error: The file 'neg.txt' in the given directory does not contain a valid never claim."

    nofStates = getNofStatesInSpinNeverClaimFile(args[0]+os.sep+"pos.txt")
except IOError:
    print "The given directory cannot be accessed. Probably it does not exist or does not contain the files 'neg.txt' or 'pos.txt'"
    exit(1)

# Perform computation using scheme "A"
for nofStates in range(1,nofStates):
    for bound in range(1,9):
        
        # Run the solver
        program = basePath+os.sep+"NBWSat"
        params = [program,args[0],str(nofStates),str(bound)]
        deadline = time.time() + timeout
        process = subprocess.Popen(args=params,executable=program,shell=False,bufsize=1024867,stderr=subprocess.PIPE,stdout=subprocess.PIPE)
        retVal = None
        while (retVal==None) and ((time.time()<deadline) or (timeout==-1)):
            retVal=process.poll();
            time.sleep(0.01)
        if (retVal==None):
            # Kill process & print best result
            if os.kill(process.pid, signal.SIGKILL)>0:
                print "Error: Cannot kill the SAT solver instance."
                exit(1)
            process.wait()
        else:
            foundResult = False
            for line in process.stdout:
                sys.stdout.write(line)
                foundResult = True    
            if foundResult:
                exit(1)
        process.stdout.close()
        process.stderr.close()
        process.wait()

# No smaller automaton found
print "No smaller automaton found!"

